from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class passivetotal(subdomain_enum_apikey):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query the PassiveTotal API for subdomains",
        "created_date": "2022-08-08",
        "author": "@TheTechromancer",
        "auth_required": True,
    }
    options = {"api_key": ""}
    options_desc = {"api_key": "PassiveTotal API Key in the format of 'username:api_key'"}

    base_url = "https://api.passivetotal.org/v2"

    async def setup(self):
        return await super().setup()

    async def ping(self):
        url = f"{self.base_url}/account/quota"
        j = (await self.api_request(url, retry_on_http_429=False)).json()
        limit = j["user"]["limits"]["search_api"]
        used = j["user"]["counts"]["search_api"]
        assert used < limit, "No quota remaining"

    def prepare_api_request(self, url, kwargs):
        api_username, api_key = self.api_key.split(":", 1)
        kwargs["auth"] = (api_username, api_key)
        return url, kwargs

    async def abort_if(self, event):
        # RiskIQ is famous for their junk data
        return await super().abort_if(event) or "unresolved" in event.tags

    async def request_url(self, query):
        url = f"{self.base_url}/enrichment/subdomains?query={self.helpers.quote(query)}"
        return await self.api_request(url)

    async def parse_results(self, r, query):
        results = set()
        for subdomain in r.json().get("subdomains", []):
            results.add(f"{subdomain}.{query}")
        return results



import logging
import regex as re
from hashlib import sha1
from radixtarget import RadixTarget
from radixtarget.helpers import host_size_key

from bbot.errors import *
from bbot.core.event import is_event
from bbot.core.event.helpers import EventSeed, BaseEventSeed
from bbot.core.helpers.misc import is_dns_name, is_ip, is_ip_type

log = logging.getLogger("bbot.core.target")


class BaseTarget(RadixTarget):
    """
    A collection of BBOT events that represent a scan target.

    The purpose of this class is to hold a potentially huge target list in a space-efficient way,
    while allowing lightning fast scope lookups.

    This class is inherited by all three components of the BBOT target:
        - Whitelist
        - Blacklist
        - Seeds
    """

    accept_target_types = ["TARGET"]

    def __init__(self, *targets, **kwargs):
        # ignore blank targets (sometimes happens as a symptom of .splitlines())
        targets = [stripped for t in targets if (stripped := (t.strip() if isinstance(t, str) else t))]
        self.event_seeds = set()
        super().__init__(*targets, **kwargs)

    @property
    def inputs(self):
        return set(e.input for e in self.event_seeds)

    def get(self, event, **kwargs):
        """
        Here we override RadixTarget's get() method, which normally only accepts hosts, to also accept events for convenience.
        """
        host = None
        raise_error = kwargs.get("raise_error", False)
        # if it's already an event or event seed, use its host
        if is_event(event) or isinstance(event, BaseEventSeed):
            host = event.host
        # save resources by checking if the event is an IP or DNS name
        elif is_ip(event, include_network=True) or is_dns_name(event):
            host = event
        # if it's a string, autodetect its type and parse out its host
        elif isinstance(event, str):
            event_seed = self._make_event_seed(event, raise_error=raise_error)
            host = event_seed.host
            if not host:
                return
        else:
            raise ValueError(f"Invalid target type for {self.__class__.__name__}: {type(event)}")
        if not host:
            msg = f"Host not found: '{event}'"
            if raise_error:
                raise KeyError(msg)
            else:
                log.warning(msg)
                return
        results = super().get(host, **kwargs)
        return results

    def _make_event_seed(self, target, raise_error=False):
        try:
            return EventSeed(target)
        except ValidationError:
            msg = f"Invalid target: '{target}'"
            if raise_error:
                raise KeyError(msg)
            else:
                log.warning(msg)

    def add(self, targets, data=None):
        if not isinstance(targets, (list, set, tuple)):
            targets = [targets]
        event_seeds = set()
        for target in targets:
            event_seed = EventSeed(target)
            if not event_seed._target_type in self.accept_target_types:
                log.warning(f"Invalid target type for {self.__class__.__name__}: {event_seed.type}")
                continue
            event_seeds.add(event_seed)

        # sort by host size to ensure consistency
        event_seeds = sorted(event_seeds, key=lambda e: ((0, 0) if not e.host else host_size_key(e.host)))
        for event_seed in event_seeds:
            self.event_seeds.add(event_seed)
            self._add(event_seed.host, data=(event_seed if data is None else data))

    def __iter__(self):
        yield from self.event_seeds


class ScanSeeds(BaseTarget):
    """
    Initial events used to seed a scan.

    These are the targets specified by the user, e.g. via `-t` on the CLI.
    """

    def get(self, event, single=True, **kwargs):
        results = super().get(event, **kwargs)
        if results and single:
            return next(iter(results))
        return results

    def _add(self, host, data):
        """
        Overrides the base method to enable having multiple events for the same host.

        The "data" attribute of the node is now a set of events.

        This is useful for seeds, because it lets us have both evilcorp.com:80 and https://evilcorp.com
            as separate events even though they have the same host.
        """
        if host:
            try:
                event_set = self.get(host, raise_error=True, single=False)
                event_set.add(data)
            except KeyError:
                event_set = {data}
            super()._add(host, data=event_set)

    def _hash_value(self):
        # seeds get hashed by event data
        return sorted(str(e.data).encode() for e in self.event_seeds)


class ACLTarget(BaseTarget):
    def __init__(self, *args, **kwargs):
        # ACL mode dedupes by host (and skips adding already-contained hosts) for efficiency
        kwargs["acl_mode"] = True
        super().__init__(*args, **kwargs)


class ScanWhitelist(ACLTarget):
    """
    A collection of BBOT events that represent a scan's whitelist.
    """

    pass


class ScanBlacklist(ACLTarget):
    """
    A collection of BBOT events that represent a scan's blacklist.
    """

    accept_target_types = ["TARGET", "BLACKLIST"]

    def __init__(self, *args, **kwargs):
        self.blacklist_regexes = set()
        super().__init__(*args, **kwargs)

    def get(self, host, **kwargs):
        """
        Blacklists only accept IPs or strings. This is cleaner since we need to search for regex patterns.
        """
        if not (is_ip_type(host) or isinstance(host, str)):
            raise ValueError(f"Invalid target type for {self.__class__.__name__}: {type(host)}")
        raise_error = kwargs.get("raise_error", False)
        # first, check event's host against blacklist
        try:
            event_seed = self._make_event_seed(host, raise_error=raise_error)
            host = event_seed.host
            to_match = event_seed.data
        except ValidationError:
            to_match = str(host)
        try:
            event_result = super().get(host, raise_error=True)
        except KeyError:
            event_result = None
        if event_result is not None:
            return event_result
        # next, check event's host against regexes
        for regex in self.blacklist_regexes:
            if regex.search(to_match):
                return host
        if raise_error:
            raise KeyError(f"Host not found: '{host}'")
        return None

    def _add(self, host, data):
        if getattr(data, "type", "") == "BLACKLIST_REGEX":
            self.blacklist_regexes.add(re.compile(data.data))
        if host is not None:
            super()._add(host, data)

    def _hash_value(self):
        # regexes are included in blacklist hash
        regex_patterns = [str(r.pattern).encode() for r in self.blacklist_regexes]
        hosts = [str(h).encode() for h in self.sorted_hosts]
        return hosts + regex_patterns

    def __len__(self):
        return super().__len__() + len(self.blacklist_regexes)

    def __bool__(self):
        return bool(len(self))


class BBOTTarget:
    """
    A convenient abstraction of a scan target that contains three subtargets:
        - seeds
        - whitelist
        - blacklist

    Provides high-level functions like in_scope(), which includes both whitelist and blacklist checks.
    """

    def __init__(self, *seeds, whitelist=None, blacklist=None, strict_scope=False):
        self.strict_scope = strict_scope
        self.seeds = ScanSeeds(*seeds, strict_dns_scope=strict_scope)
        if whitelist is None:
            whitelist = self.seeds.hosts
        self.whitelist = ScanWhitelist(*whitelist, strict_dns_scope=strict_scope)
        if blacklist is None:
            blacklist = []
        self.blacklist = ScanBlacklist(*blacklist)

    @property
    def json(self):
        return {
            "seeds": sorted(self.seeds.inputs),
            "whitelist": sorted(self.whitelist.inputs),
            "blacklist": sorted(self.blacklist.inputs),
            "strict_scope": self.strict_scope,
            "hash": self.hash.hex(),
            "seed_hash": self.seeds.hash.hex(),
            "whitelist_hash": self.whitelist.hash.hex(),
            "blacklist_hash": self.blacklist.hash.hex(),
            "scope_hash": self.scope_hash.hex(),
        }

    @property
    def hash(self):
        sha1_hash = sha1()
        for target_hash in [t.hash for t in (self.seeds, self.whitelist, self.blacklist)]:
            sha1_hash.update(target_hash)
        return sha1_hash.digest()

    @property
    def scope_hash(self):
        sha1_hash = sha1()
        # Consider only the hash values of the whitelist and blacklist
        for target_hash in [t.hash for t in (self.whitelist, self.blacklist)]:
            sha1_hash.update(target_hash)
        return sha1_hash.digest()

    def in_scope(self, host):
        """
        Check whether a hostname, url, IP, etc. is in scope.
        Accepts either events or string data.

        Checks whitelist and blacklist.
        If `host` is an event and its scope distance is zero, it will automatically be considered in-scope.

        Examples:
            Check if a URL is in scope:
            >>> preset.in_scope("http://www.evilcorp.com")
            True
        """
        blacklisted = self.blacklisted(host)
        whitelisted = self.whitelisted(host)
        return whitelisted and not blacklisted

    def blacklisted(self, host):
        """
        Check whether a hostname, url, IP, etc. is blacklisted.

        Note that `host` can be a hostname, IP address, CIDR, email address, or any BBOT `Event` with the `host` attribute.

        Args:
            host (str or IPAddress or Event): The host to check against the blacklist

        Examples:
            Check if a URL's host is blacklisted:
            >>> preset.blacklisted("http://www.evilcorp.com")
            True
        """
        return host in self.blacklist

    def whitelisted(self, host):
        """
        Check whether a hostname, url, IP, etc. is whitelisted.

        Note that `host` can be a hostname, IP address, CIDR, email address, or any BBOT `Event` with the `host` attribute.

        Args:
            host (str or IPAddress or Event): The host to check against the whitelist

        Examples:
            Check if a URL's host is whitelisted:
            >>> preset.whitelisted("http://www.evilcorp.com")
            True
        """
        return host in self.whitelist

    def __eq__(self, other):
        return self.hash == other.hash



import json
import base64
import binascii
import xmltodict
from contextlib import suppress
from urllib.parse import unquote, quote
from xml.parsers.expat import ExpatError

from bbot.core.helpers.misc import is_printable


# TODO: This logic is perfect for extracting params. We should expand it outwards to include other higher-level envelopes:
#    - QueryStringEnvelope
#    - MultipartFormEnvelope
#    - HeaderEnvelope
#    - CookieEnvelope
#
# Once we start ingesting HTTP_REQUEST events, this will make them instantly fuzzable


class EnvelopeChildTracker(type):
    """
    Keeps track of all the child envelope classes
    """

    children = []

    def __new__(mcs, name, bases, class_dict):
        # Create the class
        cls = super().__new__(mcs, name, bases, class_dict)
        # Don't register the base class itself
        if bases and not name.startswith("Base"):  # Only register if it has base classes (i.e., is a child)
            EnvelopeChildTracker.children.append(cls)
            EnvelopeChildTracker.children.sort(key=lambda x: x.priority)
        return cls


class BaseEnvelope(metaclass=EnvelopeChildTracker):
    __slots__ = ["subparams", "selected_subparam", "singleton"]

    # determines the order of the envelope detection
    priority = 5
    # whether the envelope is the final format, e.g. raw text/binary
    end_format = False
    ignore_exceptions = (Exception,)
    envelope_classes = EnvelopeChildTracker.children
    # transparent envelopes (i.e. TextEnvelope) are not counted as envelopes or included in the finding descriptions
    transparent = False

    def __init__(self, s):
        unpacked_data = self.unpack(s)

        if self.end_format:
            inner_envelope = unpacked_data
        else:
            inner_envelope = self.detect(unpacked_data)

        self.selected_subparam = None
        # if we have subparams, our inner envelope will be a dictionary
        if isinstance(inner_envelope, dict):
            self.subparams = inner_envelope
            self.singleton = False
        # otherwise if we just have one value, we make a dictionary with a default key
        else:
            self.subparams = {"__default__": inner_envelope}
            self.singleton = True

    @property
    def final_envelope(self):
        try:
            return self.unpacked_data(recursive=False).final_envelope
        except AttributeError:
            return self

    @property
    def friendly_name(self):
        if self.friendly_name:
            return self.friendly_name
        else:
            return self.name

    def pack(self, data=None):
        if data is None:
            data = self.unpacked_data(recursive=False)
            with suppress(AttributeError):
                data = data.pack()
        return self._pack(data)

    def unpack(self, s):
        return self._unpack(s)

    def _pack(self, s):
        """
        Encodes the string using the class's unique encoder (adds the outer envelope)
        """
        raise NotImplementedError("Envelope.pack() must be implemented")

    def _unpack(self, s):
        """
        Decodes the string using the class's unique encoder (removes the outer envelope)
        """
        raise NotImplementedError("Envelope.unpack() must be implemented")

    def unpacked_data(self, recursive=True):
        try:
            unpacked = self.subparams["__default__"]
            if recursive:
                with suppress(AttributeError):
                    return unpacked.unpacked_data(recursive=recursive)
            return unpacked
        except KeyError:
            return self.subparams

    @classmethod
    def detect(cls, s):
        """
        Detects the type of envelope used to encode the packed_data
        """
        if not isinstance(s, str):
            raise ValueError(f"Invalid data passed to detect(): {s} ({type(s)})")
        # if the value is empty, we just return the text envelope
        if not s.strip():
            return TextEnvelope(s)
        for envelope_class in cls.envelope_classes:
            with suppress(*envelope_class.ignore_exceptions):
                envelope = envelope_class(s)
                if envelope is not False:
                    # make sure the envelope is not just the original string, to prevent unnecessary envelope detection. For example, "10" is technically valid JSON, but nothing is being encapsulated
                    if str(envelope.unpacked_data()) == s:
                        return TextEnvelope(s)
                    else:
                        return envelope
                del envelope
        raise Exception(f"No envelope detected for data: '{s}' ({type(s)})")

    def get_subparams(self, key=None, data=None, recursive=True):
        if data is None:
            data = self.unpacked_data(recursive=recursive)
        if key is None:
            key = []

        if isinstance(data, dict):
            for k, v in data.items():
                full_key = key + [k]
                if isinstance(v, dict):
                    yield from self.get_subparams(full_key, v)
                else:
                    yield full_key, v
        else:
            yield [], data

    def get_subparam(self, key=None, recursive=True):
        if key is None:
            key = self.selected_subparam
        envelope = self
        if recursive:
            envelope = self.final_envelope
        data = envelope.unpacked_data(recursive=False)
        if key is None:
            if envelope.singleton:
                key = []
            else:
                raise ValueError("No subparam selected")
        else:
            for segment in key:
                data = data[segment]
        return data

    def set_subparam(self, key=None, value=None, recursive=True):
        envelope = self
        if recursive:
            envelope = self.final_envelope

        # if there's only one value to set, we can just set it directly
        if envelope.singleton:
            envelope.subparams["__default__"] = value
            return

        # if key isn't specified, use the selected subparam
        if key is None:
            key = self.selected_subparam
        if key is None:
            raise ValueError(f"{self} -> {envelope}: No subparam selected")

        data = envelope.unpacked_data(recursive=False)
        for segment in key[:-1]:
            data = data[segment]
        data[key[-1]] = value

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def num_envelopes(self):
        num_envelopes = 0 if self.transparent else 1
        if self.end_format:
            return num_envelopes
        for envelope in self.subparams.values():
            with suppress(AttributeError):
                num_envelopes += envelope.num_envelopes
        return num_envelopes

    @property
    def summary(self):
        if self.transparent:
            return ""
        self_string = f"{self.friendly_name}"
        with suppress(AttributeError):
            child_envelope = self.unpacked_data(recursive=False)
            child_summary = child_envelope.summary
            if child_summary:
                self_string += f" -> {child_summary}"

        if self.selected_subparam:
            self_string += f" [{'.'.join(self.selected_subparam)}]"
        return self_string

    def to_dict(self):
        return self.summary

    def __str__(self):
        return self.summary

    __repr__ = __str__


class HexEnvelope(BaseEnvelope):
    """
    Hexadecimal encoding
    """

    friendly_name = "Hexadecimal-Encoded"

    ignore_exceptions = (ValueError, UnicodeDecodeError)

    def _pack(self, s):
        return s.encode().hex()

    def _unpack(self, s):
        return bytes.fromhex(s).decode()


class B64Envelope(BaseEnvelope):
    """
    Base64 encoding
    """

    friendly_name = "Base64-Encoded"

    ignore_exceptions = (binascii.Error, UnicodeDecodeError, ValueError)

    def unpack(self, s):
        # it's easy to have a small value that accidentally decodes to base64
        if len(s) < 8 and not s.endswith("="):
            raise ValueError("Data is too small to be sure")
        return super().unpack(s)

    def _pack(self, s):
        return base64.b64encode(s.encode()).decode()

    def _unpack(self, s):
        return base64.b64decode(s).decode()


class URLEnvelope(BaseEnvelope):
    """
    URL encoding
    """

    friendly_name = "URL-Encoded"

    def unpack(self, s):
        unpacked = super().unpack(s)
        if unpacked == s:
            raise ValueError("Data is not URL-encoded")
        return unpacked

    def _pack(self, s):
        return quote(s)

    def _unpack(self, s):
        return unquote(s)


class TextEnvelope(BaseEnvelope):
    """
    Text encoding
    """

    end_format = True
    # lowest priority means text is the ultimate fallback
    priority = 10
    transparent = True
    ignore_exceptions = ()

    def _pack(self, s):
        return s

    def _unpack(self, s):
        if not is_printable(s):
            raise ValueError(f"Non-printable data detected in TextEnvelope: '{s}' ({type(s)})")
        return s


# class BinaryEnvelope(BaseEnvelope):
#     """
#     Binary encoding
#     """
#     end_format = True

#     def pack(self, s):
#         return s

#     def unpack(self, s):
#         if is_printable(s):
#             raise Exception("Non-binary data detected in BinaryEnvelope")
#         return s


class JSONEnvelope(BaseEnvelope):
    """
    JSON encoding
    """

    friendly_name = "JSON-formatted"
    end_format = True
    priority = 8
    ignore_exceptions = (json.JSONDecodeError,)

    def _pack(self, s):
        return json.dumps(s)

    def _unpack(self, s):
        return json.loads(s)


class XMLEnvelope(BaseEnvelope):
    """
    XML encoding
    """

    friendly_name = "XML-formatted"
    end_format = True
    priority = 9
    ignore_exceptions = (ExpatError,)

    def _pack(self, s):
        return xmltodict.unparse(s)

    def _unpack(self, s):
        return xmltodict.parse(s)



import random

from .base import ModuleTestBase


class TestBeVigil(ModuleTestBase):
    module_name = "bevigil"
    config_overrides = {"modules": {"bevigil": {"api_key": "asdf", "urls": True}}}

    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://osint.bevigil.com/api/blacklanternsecurity.com/subdomains/",
            match_headers={"X-Access-Token": "asdf"},
            json={
                "domain": "blacklanternsecurity.com",
                "subdomains": [
                    "asdf.blacklanternsecurity.com",
                ],
            },
        )
        module_test.httpx_mock.add_response(
            url="https://osint.bevigil.com/api/blacklanternsecurity.com/urls/",
            json={"domain": "blacklanternsecurity.com", "urls": ["https://asdf.blacklanternsecurity.com"]},
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"
        assert any(e.data == "https://asdf.blacklanternsecurity.com/" for e in events), "Failed to detect url"


class TestBeVigilMultiKey(TestBeVigil):
    api_keys = ["1234", "4321", "asdf", "fdsa"]
    random.shuffle(api_keys)
    config_overrides = {"modules": {"bevigil": {"api_key": api_keys, "urls": True}}}

    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://osint.bevigil.com/api/blacklanternsecurity.com/subdomains/",
            match_headers={"X-Access-Token": "fdsa"},
            json={
                "domain": "blacklanternsecurity.com",
                "subdomains": [
                    "asdf.blacklanternsecurity.com",
                ],
            },
        )
        module_test.httpx_mock.add_response(
            match_headers={"X-Access-Token": "asdf"},
            url="https://osint.bevigil.com/api/blacklanternsecurity.com/urls/",
            json={"domain": "blacklanternsecurity.com", "urls": ["https://asdf.blacklanternsecurity.com"]},
        )



from .base import ModuleTestBase


class TestEmailFormat(ModuleTestBase):
    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://www.email-format.com/d/blacklanternsecurity.com/",
            text="""<a href="/cdn-cgi/l/email-protection" class="__cf_email__" data-cfemail="0a63646c654a68666b6961666b647e6f7864796f697f78637e7324696567">[email&#160;protected]</a>""",
        )

    def check(self, module_test, events):
        assert any(e.data == "info@blacklanternsecurity.com" for e in events), "Failed to detect email"



from .base import ModuleTestBase


class TestWebReport(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "wappalyzer", "badsecrets", "web_report", "trufflehog"]
    config_overrides = {"modules": {"trufflehog": {"only_verified": False}}}

    async def setup_before_prep(self, module_test):
        # trufflehog --> FINDING
        # wappalyzer --> TECHNOLOGY
        # badsecrets --> VULNERABILITY
        respond_args = {"response_data": web_body}
        module_test.set_expect_requests(respond_args=respond_args)

    def check(self, module_test, events):
        report_file = module_test.scan.home / "web_report.html"
        with open(report_file) as f:
            report_content = f.read()
        assert "<li>[CRITICAL] Known Secret Found" in report_content
        assert (
            """<h3>URL</h3>
<ul>
<li><strong>http://127.0.0.1:8888/</strong>"""
            in report_content
        )
        assert """Possible Secret Found. Detector Type: [PrivateKey]""" in report_content
        assert "<h3>TECHNOLOGY</h3>" in report_content
        assert "<p>flask</p>" in report_content


web_body = """
<html>
<body>
<link rel="preconnect" href="https://fonts.googleapis.com">
<link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
<link href="https://fonts.googleapis.com/css2?family=Open+Sans+Condensed:wght@700&family=Open+Sans:ital,wght@0,400;0,600;0,700;0,800;1,400&display=swap" rel="stylesheet">
    <form method="post" action="./query.aspx" id="form1">
<div class="aspNetHidden">
<input type="hidden" name="__VIEWSTATE" id="__VIEWSTATE" value="rJdyYspajyiWEjvZ/SMXsU/1Q6Dp1XZ/19fZCABpGqWu+s7F1F/JT1s9mP9ED44fMkninhDc8eIq7IzSllZeJ9JVUME41i8ozheGunVSaESf4nBu" />
</div>

<div class="aspNetHidden">

    <input type="hidden" name="__VIEWSTATEGENERATOR" id="__VIEWSTATEGENERATOR" value="EDD8C9AE" />
    <input type="hidden" name="__VIEWSTATEENCRYPTED" id="__VIEWSTATEENCRYPTED" value="" />
</div>
    </form>
    <p>-----BEGIN PRIVATE KEY-----
MIICdwIBADANBgkqhkiG9w0BAQEFAASCAmEwggJdAgEAAoGBAOBY2pd9PSQvuxqu
WXFNVgILTWuUc721Wc2sFNvp4beowhUe1lfxaq5ZfCJcz7z4QsqFhOeks69O9UIb
oiOTDocPDog9PHO8yZXopHm0StFZvSjjKSNuFvy/WopPTGpxUZ5boCaF1CXumY7W
FL+jIap5faimLL9prIwaQKBwv80lAgMBAAECgYEAxvpHtgCgD849tqZYMgOTevCn
U/kwxltoMOClB39icNA+gxj8prc6FTTMwnVq0oGmS5UskX8k1yHCqUV1AvRU9o+q
I8L8a3F3TQKQieI/YjiUNK8A87bKkaiN65ooOnhT+I3ZjZMPR5YEyycimMp22jsv
LyX/35J/wf1rNiBs/YECQQDvtxgmMhE+PeajXqw1w2C3Jds27hI3RPDnamEyWr/L
KkSplbKTF6FuFDYOFdJNPrfxm1tx2MZ2cBfs+h/GnCJVAkEA75Z9w7q8obbqGBHW
9bpuFvLjW7bbqO7HBuXYX9zQcZL6GSArFP0ba5lhgH1qsVQfxVWVyiV9/chme7xc
ljfvkQJBAJ7MpSPQcRnRefNp6R0ok+5gFqt55PlWI1y6XS81bO7Szm+laooE0n0Q
yIpmLE3dqY9VgquVlkupkD/9poU0s40CQD118ZVAVht1/N9n1Cj9RjiE3mYspnTT
rCLM25Db6Gz6M0Y2xlaAB4S2uBhqE/Chj/TjW6WbsJJl0kRzsZynhMECQFYKiM1C
T4LB26ynW00VE8z4tEWSoYt4/Vn/5wFhalVjzoSJ8Hm2qZiObRYLQ1m0X4KnkShk
Gnl54dJHT+EhlfY=
-----END PRIVATE KEY-----</p>
</body>
</html>
"""



from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class bevigil(subdomain_enum_apikey):
    """
    Retrieve OSINT data from mobile applications using BeVigil
    """

    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME", "URL_UNVERIFIED"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Retrieve OSINT data from mobile applications using BeVigil",
        "created_date": "2022-10-26",
        "author": "@alt-glitch",
        "auth_required": True,
    }
    options = {"api_key": "", "urls": False}
    options_desc = {"api_key": "BeVigil OSINT API Key", "urls": "Emit URLs in addition to DNS_NAMEs"}

    base_url = "https://osint.bevigil.com/api"

    async def setup(self):
        self.api_key = self.config.get("api_key", "")
        self.urls = self.config.get("urls", False)
        return await super().setup()

    def prepare_api_request(self, url, kwargs):
        kwargs["headers"]["X-Access-Token"] = self.api_key
        return url, kwargs

    async def handle_event(self, event):
        query = self.make_query(event)
        subdomains = await self.query(query, request_fn=self.request_subdomains, parse_fn=self.parse_subdomains)
        if subdomains:
            for subdomain in subdomains:
                await self.emit_event(
                    subdomain,
                    "DNS_NAME",
                    parent=event,
                    context=f'{{module}} queried BeVigil\'s API for "{query}" and discovered {{event.type}}: {{event.data}}',
                )

        if self.urls:
            urls = await self.query(query, request_fn=self.request_urls, parse_fn=self.parse_urls)
            if urls:
                for parsed_url in await self.helpers.run_in_executor_mp(self.helpers.validators.collapse_urls, urls):
                    await self.emit_event(
                        parsed_url.geturl(),
                        "URL_UNVERIFIED",
                        parent=event,
                        context=f'{{module}} queried BeVigil\'s API for "{query}" and discovered {{event.type}}: {{event.data}}',
                    )

    async def request_subdomains(self, query):
        url = f"{self.base_url}/{self.helpers.quote(query)}/subdomains/"
        return await self.api_request(url)

    async def request_urls(self, query):
        url = f"{self.base_url}/{self.helpers.quote(query)}/urls/"
        return await self.api_request(url)

    async def parse_subdomains(self, r, query=None):
        results = set()
        subdomains = r.json().get("subdomains")
        if subdomains:
            results.update(subdomains)
        return results

    async def parse_urls(self, r, query=None):
        results = set()
        urls = r.json().get("urls")
        if urls:
            results.update(urls)
        return results



import time
import asyncio

from .base import ModuleTestBase


class TestPostgres(ModuleTestBase):
    targets = ["evilcorp.com"]
    skip_distro_tests = True

    async def setup_before_prep(self, module_test):
        process = await asyncio.create_subprocess_exec(
            "docker",
            "run",
            "--name",
            "bbot-test-postgres",
            "--rm",
            "-e",
            "POSTGRES_PASSWORD=bbotislife",
            "-e",
            "POSTGRES_USER=postgres",
            "-p",
            "5432:5432",
            "-d",
            "postgres",
        )

        import asyncpg

        # wait for the container to start
        start_time = time.time()
        while True:
            try:
                # Connect to the default 'postgres' database to create 'bbot'
                conn = await asyncpg.connect(
                    user="postgres", password="bbotislife", database="postgres", host="127.0.0.1"
                )
                await conn.execute("CREATE DATABASE bbot")
                await conn.close()
                break
            except asyncpg.exceptions.DuplicateDatabaseError:
                # If the database already exists, break the loop
                break
            except Exception as e:
                if time.time() - start_time > 60:  # timeout after 60 seconds
                    self.log.error("PostgreSQL server did not start in time.")
                    raise e
                await asyncio.sleep(1)

        if process.returncode != 0:
            self.log.error("Failed to start PostgreSQL server")

    async def check(self, module_test, events):
        import asyncpg

        # Connect to the PostgreSQL database
        conn = await asyncpg.connect(user="postgres", password="bbotislife", database="bbot", host="127.0.0.1")

        try:
            events = await conn.fetch("SELECT * FROM event")
            assert len(events) == 3, "No events found in PostgreSQL database"
            scans = await conn.fetch("SELECT * FROM scan")
            assert len(scans) == 1, "No scans found in PostgreSQL database"
            targets = await conn.fetch("SELECT * FROM target")
            assert len(targets) == 1, "No targets found in PostgreSQL database"
        finally:
            await conn.close()
            process = await asyncio.create_subprocess_exec(
                "docker", "stop", "bbot-test-postgres", stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
            )
            stdout, stderr = await process.communicate()

            if process.returncode != 0:
                raise Exception(f"Failed to stop PostgreSQL server: {stderr.decode()}")



from .base import ModuleTestBase


class TestGithub_Usersearch(ModuleTestBase):
    config_overrides = {"modules": {"github_usersearch": {"api_key": "asdf"}}}
    query_1 = """query search_users {
            search(query: "blacklanternsecurity.com", type: USER, first: 100, after: "") {
                userCount
                pageInfo {
                    hasNextPage
                    endCursor
                }
                edges {
                    node {
                        ... on User {
                          login
                          # bio Commented out as user can add arbritrary domains to their bio
                          email # Email is verified by github
                          websiteUrl # Website is not verified by github
                        }
                    }
                }
            }
        }"""
    query_2 = """query search_users {
            search(query: "blacklanternsecurity.com", type: USER, first: 100, after: "Y3Vyc29yOjUz") {
                userCount
                pageInfo {
                    hasNextPage
                    endCursor
                }
                edges {
                    node {
                        ... on User {
                          login
                          # bio Commented out as user can add arbritrary domains to their bio
                          email # Email is verified by github
                          websiteUrl # Website is not verified by github
                        }
                    }
                }
            }
        }"""

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(url="https://api.github.com/zen")
        module_test.httpx_mock.add_response(
            url="https://api.github.com/graphql",
            match_headers={"Authorization": "token asdf"},
            match_json={"query": self.query_1},
            json={
                "data": {
                    "search": {
                        "userCount": 2,
                        "pageInfo": {"hasNextPage": True, "endCursor": "Y3Vyc29yOjUz"},
                        "edges": [
                            {
                                "node": {
                                    "login": "user_one",
                                    "email": "test@blacklanternsecurity.com",
                                    "websiteUrl": None,
                                }
                            },
                            {"node": {"login": "user_two", "email": None, "websiteUrl": None}},
                        ],
                    }
                }
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/graphql",
            match_headers={"Authorization": "token asdf"},
            match_json={"query": self.query_2},
            json={
                "data": {
                    "search": {
                        "userCount": 1,
                        "pageInfo": {"hasNextPage": False, "endCursor": "Y3Vyc29yOjU"},
                        "edges": [
                            {
                                "node": {
                                    "login": "user_three",
                                    "email": None,
                                    "websiteUrl": "https://blog.blacklanternsecurity.com",
                                }
                            }
                        ],
                    }
                }
            },
        )

    def check(self, module_test, events):
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "github"
                and e.data["profile_name"] == "user_one"
                and str(e.module) == "github_usersearch"
                and "github-org-member" in e.tags
                and e.scope_distance == 1
            ]
        ), "Failed to find user_one github"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "EMAIL_ADDRESS"
                and e.data == "test@blacklanternsecurity.com"
                and str(e.module) == "github_usersearch"
            ]
        ), "Failed to find email address for user_one"
        assert 0 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "github"
                and e.data["profile_name"] == "user_two"
                and str(e.module) == "github_usersearch"
                and "github-org-member" in e.tags
                and e.scope_distance == 1
            ]
        ), "user_two should not be in scope due to no email or website"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "github"
                and e.data["profile_name"] == "user_three"
                and str(e.module) == "github_usersearch"
                and "github-org-member" in e.tags
                and e.scope_distance == 1
            ]
        ), "Failed to find user_three github"



from bbot.modules.templates.github import github
from bbot.modules.templates.subdomain_enum import subdomain_enum


class github_usersearch(github, subdomain_enum):
    watched_events = ["DNS_NAME"]
    produced_events = ["SOCIAL", "EMAIL_ADDRESS"]
    flags = ["passive", "safe", "code-enum"]
    meta = {
        "description": "Query Github's API for users with emails matching in scope domains that may not be discoverable by listing members of the organization.",
        "created_date": "2025-05-10",
        "author": "@domwhewell-sage",
        "auth_required": True,
    }
    options = {"api_key": ""}
    options_desc = {"api_key": "Github token"}

    async def handle_event(self, event):
        self.verbose("Searching for users with emails matching in scope domains")
        query = self.make_query(event)
        users = await self.query_users(query)
        for user, email in users:
            user_url = f"https://github.com/{user}"
            event_data = {"platform": "github", "profile_name": user, "url": user_url}
            user_event = self.make_event(event_data, "SOCIAL", tags="github-org-member", parent=event)
            if user_event:
                await self.emit_event(
                    user_event,
                    context=f"{{module}} searched for users with {{DNS_NAME}} in the profile and discovered {{event.type}}: {user_url}",
                )
            if email:
                await self.emit_event(
                    email,
                    "EMAIL_ADDRESS",
                    parent=event,
                    context=f"{{module}} found an {{event.type}} on the github profile {user_url}: {{event.data}}",
                )

    async def query_users(self, query):
        users = []
        graphql_query = f"""query search_users {{
            search(query: "{query}", type: USER, first: 100, after: "{{NEXT_KEY}}") {{
                userCount
                pageInfo {{
                    hasNextPage
                    endCursor
                }}
                edges {{
                    node {{
                        ... on User {{
                          login
                          # bio Commented out as user can add arbritrary domains to their bio
                          email # Email is verified by github
                          websiteUrl # Website is not verified by github
                        }}
                    }}
                }}
            }}
        }}"""
        async for data in self.github_graphql_request(graphql_query, "search"):
            if data:
                user_count = data.get("userCount", 0)
                self.verbose(f"Found {user_count} users with the query {query}, verifying if they are in-scope...")
                edges = data.get("edges", [])
                for node in edges:
                    user = node.get("node", {})
                    in_scope_hosts = await self.scan.extract_in_scope_hostnames(str(user))
                    if in_scope_hosts:
                        login = user.get("login", "")
                        email = user.get("email", None)
                        self.verbose(
                            f'Found in-scope hostname(s): "{in_scope_hosts}" in the profile https://github.com/{login}, the profile appears to be in-scope'
                        )
                        users.append((login, email))
        return users



import logging
from pathlib import Path
from bbot.modules.base import BaseModule


class BaseOutputModule(BaseModule):
    accept_dupes = True
    _type = "output"
    scope_distance_modifier = None
    _stats_exclude = True
    _shuffle_incoming_queue = False

    def human_event_str(self, event):
        event_type = f"[{event.type}]"
        event_tags = ""
        if getattr(event, "tags", []):
            event_tags = f"\t({', '.join(sorted(getattr(event, 'tags', [])))})"
        event_str = f"{event_type:<20}\t{event.data_human}\t{event.module_sequence}{event_tags}"
        return event_str

    def _event_precheck(self, event):
        reason = "precheck succeeded"
        # special signal event types
        if event.type in ("FINISHED",):
            return True, "its type is FINISHED"
        if self.errored:
            return False, "module is in error state"
        # exclude non-watched types
        if not any(t in self.get_watched_events() for t in ("*", event.type)):
            return False, "its type is not in watched_events"
        if self.target_only:
            if "target" not in event.tags:
                return False, "it did not meet target_only filter criteria"

        ### begin output-module specific ###

        # force-output certain events to the graph
        if self._is_graph_important(event):
            return True, "event is critical to the graph"

        # omit certain event types
        if event._omit:
            if "target" in event.tags:
                reason = "it's a target"
                self.debug(f"Allowing omitted event: {event} because {reason}")
            elif event.type in self.get_watched_events():
                reason = "its type is explicitly in watched_events"
                self.debug(f"Allowing omitted event: {event} because {reason}")
            else:
                return False, "_omit is True"

        # internal events like those from speculate, ipneighbor
        # or events that are over our report distance
        if event._internal:
            return False, "_internal is True"

        return True, reason

    async def _event_postcheck(self, event):
        acceptable, reason = await super()._event_postcheck(event)
        if acceptable and not event._stats_recorded and event.type not in ("FINISHED",):
            event._stats_recorded = True
            self.scan.stats.event_produced(event)
        return acceptable, reason

    def is_incoming_duplicate(self, event, add=False):
        is_incoming_duplicate, reason = super().is_incoming_duplicate(event, add=add)
        # make exception for graph-important events
        if self._is_graph_important(event):
            return False, "event is graph-important"
        return is_incoming_duplicate, reason

    def _prep_output_dir(self, filename):
        self.output_file = self.config.get("output_file", "")
        if self.output_file:
            self.output_file = Path(self.output_file)
        else:
            self.output_file = self.scan.home / str(filename)
        self.helpers.mkdir(self.output_file.parent)
        self._file = None

    def _scope_distance_check(self, event):
        return True, ""

    @property
    def file(self):
        if getattr(self, "_file", None) is None:
            self._file = open(self.output_file, mode="a")
        return self._file

    @property
    def log(self):
        if self._log is None:
            self._log = logging.getLogger(f"bbot.modules.output.{self.name}")
        return self._log



from .base import ModuleTestBase


class TestDehashed(ModuleTestBase):
    modules_overrides = ["dehashed", "speculate"]
    config_overrides = {
        "scope": {"report_distance": 2},
        "modules": {"dehashed": {"api_key": "deadbeef"}},
    }

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.dehashed.com/v2/search",
            method="POST",
            json={
                "balance": 10000,
                "entries": [
                    {
                        "id": "4363462346",
                        "email": ["bob@blacklanternsecurity.com"],
                        "ip_address": ["127.0.0.9"],
                        "username": ["bob@bob.com"],
                        "hashed_password": ["$2a$12$pVmwJ7pXEr3mE.DmCCE4fOUDdeadbeefd2KuCy/tq1ZUFyEOH2bve"],
                        "name": ["Bob Smith"],
                        "phone": ["+91283423839"],
                        "database_name": "eatstreet",
                        "raw_record": {"le_only": True, "unstructured": True},
                    },
                    {
                        "id": "234623453454",
                        "email": ["tim@blacklanternsecurity.com"],
                        "username": ["timmy"],
                        "password": ["TimTamSlam69"],
                        "name": "Tim Tam",
                        "phone": ["+123455667"],
                        "database_name": "eatstreet",
                    },
                ],
                "took": "61ms",
                "total": 2,
            },
        )
        await module_test.mock_dns(
            {
                "bob.com": {"A": ["127.0.0.1"]},
                "blacklanternsecurity.com": {"A": ["127.0.0.1"]},
            }
        )

    def check(self, module_test, events):
        assert len(events) == 12
        assert 1 == len([e for e in events if e.type == "DNS_NAME" and e.data == "blacklanternsecurity.com"])
        assert 1 == len([e for e in events if e.type == "ORG_STUB" and e.data == "blacklanternsecurity"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "EMAIL_ADDRESS"
                and e.data == "bob@bob.com"
                and e.scope_distance == 1
                and "affiliate" in e.tags
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME" and e.data == "bob.com" and e.scope_distance == 1 and "affiliate" in e.tags
            ]
        )
        assert 1 == len([e for e in events if e.type == "EMAIL_ADDRESS" and e.data == "bob@blacklanternsecurity.com"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "USERNAME"
                and e.data == "bob@blacklanternsecurity.com:bob@bob.com"
                and e.parent.data == "bob@blacklanternsecurity.com"
            ]
        )
        assert 1 == len([e for e in events if e.type == "EMAIL_ADDRESS" and e.data == "tim@blacklanternsecurity.com"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "HASHED_PASSWORD"
                and e.data
                == "bob@blacklanternsecurity.com:$2a$12$pVmwJ7pXEr3mE.DmCCE4fOUDdeadbeefd2KuCy/tq1ZUFyEOH2bve"
            ]
        )
        assert 1 == len(
            [e for e in events if e.type == "PASSWORD" and e.data == "tim@blacklanternsecurity.com:TimTamSlam69"]
        )
        assert 1 == len([e for e in events if e.type == "USERNAME" and e.data == "tim@blacklanternsecurity.com:timmy"])


class TestDehashedBadEmail(TestDehashed):
    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.dehashed.com/v2/search",
            method="POST",
            json={
                "balance": 10000,
                "entries": [
                    {
                        "id": "EZxg4Lz-INLUt6uRXZaV",
                        "email": ["foo.example.com"],
                        "database_name": "Collections",
                    },
                ],
                "took": "41ms",
                "total": 1,
            },
        )

    def check(self, module_test, events):
        debug_log_content = open(module_test.scan.home / "debug.log").read()
        assert "Invalid email from dehashed.com: foo.example.com" in debug_log_content


class TestDehashedHTTPError(TestDehashed):
    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.dehashed.com/v2/search",
            method="POST",
            json={"error": "issue with request body"},
            status_code=400,
        )

    def check(self, module_test, events):
        scan_log_content = open(module_test.scan.home / "scan.log").read()
        assert (
            'Error retrieving results from dehashed.com (status code 400): {"error":"issue with request body"}'
            in scan_log_content
        )


class TestDehashedTooManyResults(TestDehashed):
    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.dehashed.com/v2/search",
            method="POST",
            json={
                "balance": 10000,
                "entries": [
                    {
                        "id": "VXhNxj46SGsW4Lworh-G",
                        "email": ["bob@bob.com"],
                        "database_name": "Collections",
                    },
                ],
                "took": "40ms",
                "total": 10001,
            },
        )

    def check(self, module_test, events):
        scan_log_content = open(module_test.scan.home / "scan.log").read()
        assert "has 10,001 results in Dehashed. The API can only process the first 10,000 results." in scan_log_content



import re
from .base import ModuleTestBase


class TestTelerik(ModuleTestBase):
    targets = ["http://127.0.0.1:8888", "http://127.0.0.1:8888/telerik.aspx"]
    modules_overrides = ["httpx", "telerik"]
    config_overrides = {"modules": {"telerik": {"exploit_RAU_crypto": True}}}

    async def setup_before_prep(self, module_test):
        # Simulate Telerik.Web.UI.WebResource.axd?type=rau detection
        expect_args = {"method": "GET", "uri": "/Telerik.Web.UI.WebResource.axd", "query_string": "type=rau"}
        respond_args = {
            "response_data": '{ "message" : "RadAsyncUpload handler is registered succesfully, however, it may not be accessed directly." }'
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate Vulnerable Telerik.Web.UI.WebResource.axd
        vuln_data = "ATTu5i4R+ViNFYO6kst0jC11wM/1iqH+W/isjhaDjNuCI7eJ/BY5d1E9eqZK27CJCMuon9u8/hgRIM/cTlgLlv4qOYjPBjs81Y3dAZAdtIr3TXiCmZi9M09a1BYMxjvGKfVky3b7PoOppeWS/3rglTwL1e8oyqLGx2NKUH5y8Cd+kLKV2f31J1sV4I5HTDKgDmvziJp3zlDrCb0Fi9ilKH+O1cbVx6SdBop/U30FxLaB/QIbt2N1rQHREJ5Skpgo7dilPxzBaTObdBhCVyB/FiJhenS/0u3h0Mpi6+A40SylICcyyxQha7+Uh7lEJ8Ne+2eTs4WqcaaQbvIhy7oHc+D0soxRKMZRjo7Up+UWHQJJh6KtWSCxUESNSdNcxjPQZE9HqsPlldVlkeC+ehSGce5bR0Ylots6Iz1OoCgMEWwxByeG3VzgxF6XpitL61A1hFcNo9euSTnCfOWh0vrQHON7DN5LpM9xr7SoD0Dnu01hZ9NS1PHhPLyN5WS87u5qdZp/z3Sxwc3wawIdo62RNf4Iz2gAKJZnPfxrE1mRn5kBe7f6O44rcuv6lcdao/DGlwbERKwRI6/n+FxGmc7H5iEKyihIwS2XUoOgsYTx5CWCDM8CuOXTk+H5fPYp9APRPbkD1IS9I/vRmvNPwWsgv8/7DzttqdBsGxiZJfCw1uZ7KSVmbItgXPAcscNxGEMaHXyJzkAl/mlM5/t/YSejwYoSW6jFfQcLdaVx2dpIpl5UmmQjFedzKeiNqpZDCk4yzXFHX24XUODYMJDtIJK2Hz1KTZmFG+LAOJjB9QOI58hFAnytcKay+JWFrzah/IvoNZxJUtlYdxw0YEyKs/ExET7AXgYQN0S+8j2PfaMMpzDSctTqpp5XBFV4Mt718GiqVnQJtWQv2p9Xl8XXOerBthbzzAciVcB8AV2WfZ51W3e4aX4kcyT/sCJhm7NR5WrNG5mX/ns0TTnGnzlPYhJcbu8uMFjMGDpXuhVyroJ7wmZucaIvesg0h5Y9cMEFviqsdy15vjMzFh+v9uO9Vicf6n9Z9JGSpWKE8wer2JU5b53Zw0cTfulAAffLWXnzOnfu&6R/cGaqQeHVAzdJ9wTFOyCsrMSTtqcjLe8AHwiPckPDUwecnJyNlkDYwDQpxGYQ9hs6YxhupK310sbCbtXB4H6Dz5rGNL40nkkyo4j2clmRr08jtFsPQ0RpE5BGsulPT3l0MxyAvPFMs8bMybUyAP+9RB9LoHE3Xo8BqDadX3HQakpPfGtiDMp+wxkWRgaNpCnXeY1QewWTF6z/duLzbu6CT6s+H4HgBHrOLTpemC2PvP2bDm0ySPHLdpapLYxU8nIYjLKIyYJgwv9S9jNckIVpcGVTWVul7CauCKxAB2mMnM9jJi8zfFwKajT5d2d9XfpkiVMrdlmikSB/ehyX1wQ=="
        expect_args = {
            "method": "POST",
            "uri": "/Telerik.Web.UI.WebResource.axd",
            "query_string": "type=rau",
            "data": vuln_data,
        }
        respond_args = {
            "response_data": '{"fileInfo":{"FileName":"RAU_crypto.bypass","ContentType":"text/html","ContentLength":5,"DateJson":"2019-01-02T03:04:05.067Z","Index":0}, "metaData":"CS8S/Z0J/b2982DRxDin0BBslA7fI0cWMuWlPu4W3FkE4tKaVoIEiAOtVlJ6D+0RQsfu8ox6gvMYxceQ0LtWyTkQBaIUa8LgLQg05DMaQuufHNx0YQ2ACi5neqDBvduj2MGiSGC0hNKzSWsHystZGUfFPLTZuJXYnff+WXurecuRzSI7d4Q1aj0bcTKKvfyQtH+fsTEafWRRZ99X/xgi4ON2OsRZ738uQHw7pQT2e1v7AtN46mxO/BmhEuZQr6m6HEvxK0pJRNkBhFUiQ+poeu8j3JzicOjvPDwFE4Rjqf3RVILt83XZrju2VpRIJqAEtf//znhH8BhT5BWvhnRo+J3ML5qoZLa2joE/QK8Ctf3UPvAFkHIUMdOH2mLNgZ+U87tdVE6fYfzvphZsLxmJRG45H8ZTZuYhJbOfei2LQ4fqHmr7p8KpJNVqoz/ev1dnBclAf5ayb40qJKEVsGXIbWEbIZwg7TTsLFc29aP7DPg=" }'
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate SpellCheckHandler detection
        expect_args = {"method": "GET", "uri": "/Telerik.Web.UI.SpellCheckHandler.axd"}
        respond_args = {"status": 500}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate SpellCheckHandler false positive detection
        expect_args = {"method": "GET", "uri": "/AAAAAAAAAAAAAA.axd"}
        respond_args = {"status": 200}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate DialogHandler detection
        expect_args = {"method": "GET", "uri": "/App_Master/Telerik.Web.UI.DialogHandler.aspx"}
        respond_args = {
            "response_data": '<input type="hidden" name="dialogParametersHolder" id="dialogParametersHolder" /><div style=\'color:red\'>Cannot deserialize dialog parameters. Please refresh the editor page.</div><div>Error Message:Invalid length for a Base-64 char array or string.</div></form></body></html>'
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate ChartImage.axd Detection
        expect_args = {
            "method": "GET",
            "uri": "/ChartImage.axd",
            "query_string": "ImageName=bqYXJAqm315eEd6b%2bY4%2bGqZpe7a1kY0e89gfXli%2bjFw%3d",
        }
        respond_args = {"status": 200}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/ChartImage.axd", "query_string": "ImageName="}
        respond_args = {"status": 500}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate Dialog Parameters in URL
        expect_args = {"method": "GET", "uri": "/telerik.aspx"}
        respond_args = {"response_data": '{"ImageManager":{"SerializedParameters":"MBwZB"}'}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Fallback
        expect_args = {"uri": re.compile(r"^/\w{10}$")}
        respond_args = {"status": 200}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["telerik"].helpers.rand_string = lambda *args, **kwargs: "AAAAAAAAAAAAAA"
        module_test.scan.modules["telerik"].telerikVersions = ["2014.2.724", "2014.3.1024", "2015.1.204"]
        module_test.scan.modules["telerik"].DialogHandlerUrls = [
            "Admin/ServerSide/Telerik.Web.UI.DialogHandler.aspx",
            "App_Master/Telerik.Web.UI.DialogHandler.aspx",
            "AsiCommon/Controls/ContentManagement/ContentDesigner/Telerik.Web.UI.DialogHandler.aspx",
        ]

    def check(self, module_test, events):
        telerik_axd_detection = False
        telerik_axd_vulnerable = False
        telerik_spellcheck_detection = False
        telerik_dialoghandler_detection = False
        telerik_chartimage_detection = False
        telerik_http_response_parameters_detection = False

        for e in events:
            if e.type == "FINDING" and "Telerik RAU AXD Handler detected" in e.data["description"]:
                e.data["description"]
                telerik_axd_detection = True
                continue

            if e.type == "VULNERABILITY" and "Confirmed Vulnerable Telerik (version: 2014.3.1024)":
                telerik_axd_vulnerable = True
                continue

            if e.type == "FINDING" and "Telerik DialogHandler detected" in e.data["description"]:
                telerik_dialoghandler_detection = True
                continue

            if e.type == "FINDING" and "Telerik SpellCheckHandler detected" in e.data["description"]:
                telerik_spellcheck_detection = True
                continue

            if e.type == "FINDING" and "Telerik ChartImage AXD Handler Detected" in e.data["description"]:
                telerik_chartimage_detection = True
                continue

            if (
                e.type == "FINDING"
                and "Telerik DialogHandler [SerializedParameters] Detected in HTTP Response" in e.data["description"]
            ):
                telerik_http_response_parameters_detection = True
                continue

        assert telerik_axd_detection, "Telerik AXD detection failed"
        assert telerik_axd_vulnerable, "Telerik vulnerable AXD detection failed"
        assert telerik_spellcheck_detection, "Telerik spellcheck detection failed"
        assert telerik_dialoghandler_detection, "Telerik dialoghandler detection failed"
        assert telerik_chartimage_detection, "Telerik chartimage detection failed"
        assert telerik_http_response_parameters_detection, "Telerik SerializedParameters detection failed"


class TestTelerikDialogHandler_includesubdirs(TestTelerik):
    targets = ["http://127.0.0.1:8888/", "http://127.0.0.1:8888/temp/"]
    config_overrides = {
        "modules": {
            "telerik": {
                "include_subdirs": True,
            },
        }
    }
    modules_overrides = ["httpx", "telerik"]

    async def setup_before_prep(self, module_test):
        # Simulate NO SpellCheckHandler detection (not testing for that with this test)
        expect_args = {"method": "GET", "uri": "/Telerik.Web.UI.SpellCheckHandler.axd"}
        respond_args = {"status": 404}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate DialogHandler detection
        expect_args = {"method": "GET", "uri": "/App_Master/Telerik.Web.UI.DialogHandler.aspx"}
        respond_args = {
            "response_data": '<input type="hidden" name="dialogParametersHolder" id="dialogParametersHolder" /><div style=\'color:red\'>Cannot deserialize dialog parameters. Please refresh the editor page.</div><div>Error Message:Invalid length for a Base-64 char array or string.</div></form></body></html>'
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate DialogHandler detection (in /temp)
        expect_args = {"method": "GET", "uri": "/temp/App_Master/Telerik.Web.UI.DialogHandler.aspx"}
        respond_args = {
            "response_data": '<input type="hidden" name="dialogParametersHolder" id="dialogParametersHolder" /><div style=\'color:red\'>Cannot deserialize dialog parameters. Please refresh the editor page.</div><div>Error Message:Invalid length for a Base-64 char array or string.</div></form></body></html>'
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Simulate /temp directory detection
        expect_args = {"method": "GET", "uri": "/temp/"}
        respond_args = {"response_data": "Temporary directory found"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        # Fallback
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["telerik"].telerikVersions = ["2014.2.724", "2014.3.1024", "2015.1.204"]
        module_test.scan.modules["telerik"].DialogHandlerUrls = [
            "App_Master/Telerik.Web.UI.DialogHandler.aspx",
        ]

    def check(self, module_test, events):
        # Check if the expected requests were made
        finding_count = sum(
            1 for e in events if e.type == "FINDING" and "Telerik DialogHandler detected" in e.data["description"]
        )
        assert finding_count == 2, "Expected 2 FINDING events (root and /temp), got {finding_count}"



import json
from contextlib import suppress

from bbot.modules.output.base import BaseOutputModule


class JSON(BaseOutputModule):
    watched_events = ["*"]
    meta = {
        "description": "Output to Newline-Delimited JSON (NDJSON)",
        "created_date": "2022-04-07",
        "author": "@TheTechromancer",
    }
    options = {"output_file": "", "siem_friendly": False}
    options_desc = {
        "output_file": "Output to file",
        "siem_friendly": "Output JSON in a SIEM-friendly format for ingestion into Elastic, Splunk, etc.",
    }
    _preserve_graph = True

    async def setup(self):
        self._prep_output_dir("output.json")
        self.siem_friendly = self.config.get("siem_friendly", False)
        return True

    async def handle_event(self, event):
        event_json = event.json(siem_friendly=self.siem_friendly)
        event_str = json.dumps(event_json)
        if self.file is not None:
            self.file.write(event_str + "\n")
            self.file.flush()

    async def cleanup(self):
        if getattr(self, "_file", None) is not None:
            with suppress(Exception):
                self.file.close()

    async def report(self):
        if self._file is not None:
            self.info(f"Saved JSON output to {self.output_file}")



from pathlib import Path
from subprocess import CalledProcessError
from bbot.modules.templates.github import github


class git_clone(github):
    watched_events = ["CODE_REPOSITORY"]
    produced_events = ["FILESYSTEM"]
    flags = ["passive", "safe", "slow", "code-enum"]
    meta = {
        "description": "Clone code github repositories",
        "created_date": "2024-03-08",
        "author": "@domwhewell-sage",
    }
    options = {"api_key": "", "output_folder": ""}
    options_desc = {
        "api_key": "Github token",
        "output_folder": "Folder to clone repositories to. If not specified, cloned repositories will be deleted when the scan completes, to minimize disk usage.",
    }

    deps_apt = ["git"]

    scope_distance_modifier = 2

    async def setup(self):
        output_folder = self.config.get("output_folder")
        self.output_dir = Path(output_folder) / "git_repos" if output_folder else self.scan.temp_dir / "git_repos"
        self.helpers.mkdir(self.output_dir)
        return await super().setup()

    async def filter_event(self, event):
        if event.type == "CODE_REPOSITORY" and "git" not in event.tags:
            return False, "event is not a git repository"
        return True

    async def handle_event(self, event):
        repository_url = event.data.get("url")
        repository_path = await self.clone_git_repository(repository_url)
        if repository_path:
            self.verbose(f"Cloned {repository_url} to {repository_path}")
            codebase_event = self.make_event({"path": str(repository_path)}, "FILESYSTEM", tags=["git"], parent=event)
            await self.emit_event(
                codebase_event,
                context=f"{{module}} cloned git repository at {repository_url} to {{event.type}}: {repository_path}",
            )

    async def clone_git_repository(self, repository_url):
        owner = repository_url.split("/")[-2]
        folder = self.output_dir / owner
        self.helpers.mkdir(folder)

        command = ["git", "-C", folder, "clone", repository_url]
        env = {"GIT_TERMINAL_PROMPT": "0"}

        try:
            hostname = self.helpers.urlparse(repository_url).hostname
            if hostname and self.api_key:
                _, domain = self.helpers.split_domain(hostname)
                # only use the api key if the domain is github.com
                if domain == "github.com":
                    env["GIT_HELPER"] = (
                        f'!f() {{ case "$1" in get) '
                        f"echo username=x-access-token; "
                        f"echo password={self.api_key};; "
                        f'esac; }}; f "$@"'
                    )
                    command = (
                        command[:1]
                        + [
                            "-c",
                            "credential.helper=",
                            "-c",
                            "credential.useHttpPath=true",
                            "--config-env=credential.helper=GIT_HELPER",
                        ]
                        + command[1:]
                    )

            output = await self.run_process(command, env=env, check=True)
        except CalledProcessError as e:
            self.debug(f"Error cloning {repository_url}. STDERR: {repr(e.stderr)}")
            return

        folder_name = output.stderr.split("Cloning into '")[1].split("'")[0]
        repo_folder = folder / folder_name

        # sanitize the repo
        # this moves the git config, index file, and hooks folder out of the .git folder to prevent nasty things
        # Note: the index file can be regenerated by running "git checkout HEAD -- ."
        self.helpers.sanitize_git_repo(repo_folder)

        return repo_folder



import logging

from bbot.core.helpers.regexes import dns_name_extraction_regex
from bbot.core.helpers.misc import clean_dns_record, smart_decode

log = logging.getLogger("bbot.core.helpers.dns")


# the following are the result of a 1-day internet survey to find the top SRV records
# the scan resulted in 36,282 SRV records. the count for each one is shown.
common_srvs = [
    "_sipfederationtls._tcp",  # 6909
    "_sip._tls",  # 6853
    "_autodiscover._tcp",  # 4268
    "_xmpp-server._tcp",  # 1437
    "_sip._tcp",  # 1193
    "_sips._tcp",  # 1183
    "_caldavs._tcp",  # 1179
    "_carddavs._tcp",  # 1132
    "_caldav._tcp",  # 1035
    "_carddav._tcp",  # 1024
    "_sip._udp",  # 1007
    "_imaps._tcp",  # 1007
    "_submission._tcp",  # 906
    "_h323cs._tcp",  # 846
    "_h323ls._udp",  # 782
    "_xmpp-client._tcp",  # 689
    "_pop3s._tcp",  # 394
    "_jabber._tcp",  # 277
    "_imap._tcp",  # 267
    "_turn._udp",  # 256
    "_pop3._tcp",  # 221
    "_ldap._tcp",  # 213
    "_smtps._tcp",  # 195
    "_sipinternaltls._tcp",  # 192
    "_vlmcs._tcp",  # 165
    "_kerberos._udp",  # 163
    "_kerberos._tcp",  # 148
    "_kpasswd._udp",  # 128
    "_kpasswd._tcp",  # 100
    "_ntp._udp",  # 90
    "_gc._tcp",  # 73
    "_kerberos-master._udp",  # 66
    "_ldap._tcp.dc._msdcs",  # 63
    "_matrix._tcp",  # 62
    "_smtp._tcp",  # 61
    "_stun._udp",  # 57
    "_kerberos._tcp.dc._msdcs",  # 54
    "_ldap._tcp.gc._msdcs",  # 49
    "_kerberos-adm._tcp",  # 44
    "_ldap._tcp.pdc._msdcs",  # 43
    "_kerberos-master._tcp",  # 43
    "_http._tcp",  # 37
    "_h323rs._tcp",  # 36
    "_sipinternal._tcp",  # 35
    "_turn._tcp",  # 33
    "_stun._tcp",  # 33
    "_h323ls._tcp",  # 33
    "_x-puppet._tcp",  # 30
    "_h323cs._udp",  # 27
    "_stuns._tcp",  # 26
    "_jabber-client._tcp",  # 25
    "_x-puppet-ca._tcp",  # 22
    "_ts3._udp",  # 22
    "_minecraft._tcp",  # 22
    "_turns._tcp",  # 21
    "_ldaps._tcp",  # 21
    "_xmpps-client._tcp",  # 20
    "_https._tcp",  # 19
    "_ftp._tcp",  # 19
    "_xmpp-server._udp",  # 18
    "_xmpp-client._udp",  # 17
    "_jabber._udp",  # 17
    "_jabber-client._udp",  # 17
    "_xmpps-server._tcp",  # 15
    "_finger._tcp",  # 14
    "_stuns._udp",  # 12
    "_hkp._tcp",  # 12
    "_vlmcs._udp",  # 11
    "_turns._udp",  # 11
    "_tftp._udp",  # 11
    "_ssh._tcp",  # 11
    "_rtps._udp",  # 11
    "_mysqlsrv._tcp",  # 11
    "_hkps._tcp",  # 11
    "_h323be._udp",  # 11
    "_dns._tcp",  # 11
    "_wss._tcp",  # 10
    "_wpad._tcp",  # 10
    "_whois._tcp",  # 10
    "_webexconnect._tcp",  # 10
    "_webexconnects._tcp",  # 10
    "_vnc._tcp",  # 10
    "_test._tcp",  # 10
    "_telnet._tcp",  # 10
    "_telnets._tcp",  # 10
    "_teamspeak._tcp",  # 10
    "_svns._tcp",  # 10
    "_svcp._tcp",  # 10
    "_smb._tcp",  # 10
    "_sip-tls._tcp",  # 10
    "_sftp._tcp",  # 10
    "_secure-pop3._tcp",  # 10
    "_secure-imap._tcp",  # 10
    "_rtsp._tcp",  # 10
    "_rtps._tcp",  # 10
    "_rpc._tcp",  # 10
    "_rfb._tcp",  # 10
    "_raop._tcp",  # 10
    "_pstn._tcp",  # 10
    "_presence._tcp",  # 10
    "_pkixrep._tcp",  # 10
    "_pgprevokations._tcp",  # 10
    "_pgpkeys._tcp",  # 10
    "_ocsp._tcp",  # 10
    "_nntp._tcp",  # 10
    "_nfs._tcp",  # 10
    "_netbios-ssn._tcp",  # 10
    "_netbios-ns._tcp",  # 10
    "_netbios-dgm._tcp",  # 10
    "_mumble._tcp",  # 10
    "_msrpc._tcp",  # 10
    "_mqtts._tcp",  # 10
    "_minecraft._udp",  # 10
    "_iscsi._tcp",  # 10
    "_ircs._tcp",  # 10
    "_ipp._tcp",  # 10
    "_ipps._tcp",  # 10
    "_h323be._tcp",  # 10
    "_gits._tcp",  # 10
    "_ftps._tcp",  # 10
    "_ftpes._tcp",  # 10
    "_dnss._udp",  # 10
    "_dnss._tcp",  # 10
    "_diameter._tcp",  # 10
    "_crl._tcp",  # 10
    "_crls._tcp",  # 10
    "_cmp._tcp",  # 10
    "_certificates._tcp",  # 10
    "_aix._tcp",  # 10
    "_afpovertcp._tcp",  # 10
    "_collab-edge._tls",  # 6
    "_tcp",  # 5
    "_client._smtp",  # 3
    "_udp",  # 2
    "_tls",  # 2
    "_msdcs",  # 2
    "_gc._msdcs",  # 2
    "_ldaps._tcp.dc._msdcs",  # 1
    "_kerberos._tcp.kdc._msdcs",  # 1
    "_kerberos.tcp.dc._msdcs",  # 1
    "_imap",  # 1
    "_iax",  # 1
]


def extract_targets(record):
    """
    Extracts hostnames or IP addresses from a given DNS record.

    This method reads the DNS record's type and based on that, extracts the target
    hostnames or IP addresses it points to. The type of DNS record
    (e.g., "A", "MX", "CNAME", etc.) determines which fields are used for extraction.

    Args:
        record (dns.rdata.Rdata): The DNS record to extract information from.

    Returns:
        set: A set of tuples, each containing the DNS record type and the extracted value.

    Examples:
        >>> from dns.rrset import from_text
        >>> record = from_text('www.example.com', 3600, 'IN', 'A', '192.0.2.1')
        >>> extract_targets(record[0])
        {('A', '192.0.2.1')}

        >>> record = from_text('example.com', 3600, 'IN', 'MX', '10 mail.example.com.')
        >>> extract_targets(record[0])
        {('MX', 'mail.example.com')}

    """
    results = set()

    def add_result(rdtype, _record):
        cleaned = clean_dns_record(_record)
        if cleaned:
            results.add((rdtype, cleaned))

    rdtype = str(record.rdtype.name).upper()
    if rdtype in ("A", "AAAA", "NS", "CNAME", "PTR"):
        add_result(rdtype, record)
    elif rdtype == "SOA":
        add_result(rdtype, record.mname)
    elif rdtype == "MX":
        add_result(rdtype, record.exchange)
    elif rdtype == "SRV":
        add_result(rdtype, record.target)
    elif rdtype == "TXT":
        for s in record.strings:
            s = smart_decode(s)
            for match in dns_name_extraction_regex.finditer(s):
                start, end = match.span()
                host = s[start:end]
                add_result(rdtype, host)
    elif rdtype == "NSEC":
        add_result(rdtype, record.next)
    else:
        log.warning(f'Unknown DNS record type "{rdtype}"')
    return results


def service_record(host, rdtype=None):
    """
    Indicates that the provided host name and optional rdtype is an SRV or related service record.

    These types of records do/should not have A/AAAA/CNAME or similar records, and are simply used to advertise configuration information and/or policy information for different Internet facing services.

    This function exists to provide a consistent way in which to perform this test, rather than having duplicated code in multiple places in different modules.

    The response provides a way for modules to quickly test whether a host name is relevant and worth inspecting or using in context of what the module does.

    NOTE: While underscores are technically not supposed to exist in DNS names as per RFC's, they can be used, so we can't assume that a name that contains or starts with an underscore is a service record and so must check for specific strings.

    Args:
        host (string): A DNS host name

    Returns:
        bool: A boolean, True indicates that the host is an SRV or similar record, False indicates that it is not.

    Examples:
        >>> service_record('_xmpp._tcp.example.com')
        True

        >>> service_record('_custom._service.example.com', 'SRV')
        True

        >>> service_record('_dmarc.example.com')
        True

        >>> service_record('www.example.com')
        False
    """

    # if we were providing an rdtype, check if it is SRV
    # NOTE: we don't care what the name is if rdtype == SRV
    if rdtype and str(rdtype).upper() == "SRV":
        return True

    # we did not receive rdtype, so we'll have to inspect host name parts
    parts = str(host).split(".")

    if not parts:
        return False

    # DMARC TXT records, e.g. _dmarc.example.com
    if parts[0] == "_dmarc":
        return True

    # MTA-STS TXT records, e.g. _mta-sts.example.com
    if parts[0] == "_mta-sts":
        return True

    if len(parts) < 2:
        return False

    # classic SRV record names, e.g. _ldap._tcp.example.com
    if parts[1] == "_udp" or parts[1] == "_tcp":
        return True

    # TLS indicating records, used by SMTP TLS-RPT etc, e.g. _smtp._tls.example.com
    if parts[1] == "_tls":
        return True

    # BIMI TXT records, e.g. selector._bimi.example.com
    if parts[1] == "_bimi":
        return True

    # DKIM TXT records, e.g. selector._domainkey.example.com
    if parts[1] == "_domainkey":
        return True

    return False



import os
import asyncio
import logging
import traceback
from signal import SIGINT
from subprocess import CompletedProcess, CalledProcessError, SubprocessError

from .misc import smart_decode, smart_encode, which

log = logging.getLogger("bbot.core.helpers.command")


async def run(self, *command, check=False, text=True, idle_timeout=None, **kwargs):
    """Runs a command asynchronously and gets its output as a string.

    This method is a simple helper for executing a command and capturing its output.
    If an error occurs during execution, it can optionally raise an error or just log the stderr.

    Args:
        *command (str): The command to run as separate arguments.
        check (bool, optional): If set to True, raises an error if the subprocess exits with a non-zero status.
                                Defaults to False.
        text (bool, optional): If set to True, decodes the subprocess output to string. Defaults to True.
        idle_timeout (int, optional): Sets a limit on the number of seconds the process can run before throwing a TimeoutError
        **kwargs (dict): Additional keyword arguments for the subprocess.

    Returns:
        CompletedProcess: A completed process object with attributes for the command, return code, stdout, and stderr.

    Raises:
        CalledProcessError: If the subprocess exits with a non-zero status and `check=True`.

    Examples:
        >>> process = await run(["ls", "/tmp"])
        >>> process.stdout
        "file1.txt\nfile2.txt"
    """
    # proc_tracker optionally keeps track of which processes are running under which modules
    # this allows for graceful SIGINTing of a module's processes in the case when it's killed
    proc_tracker = kwargs.pop("_proc_tracker", set())
    log_stderr = kwargs.pop("_log_stderr", True)
    proc, _input, command = await self._spawn_proc(*command, **kwargs)
    if proc is not None:
        proc_tracker.add(proc)
        try:
            if _input is not None:
                if isinstance(_input, (list, tuple)):
                    _input = b"\n".join(smart_encode(i) for i in _input) + b"\n"
                else:
                    _input = smart_encode(_input)

            try:
                if idle_timeout is not None:
                    stdout, stderr = await asyncio.wait_for(proc.communicate(_input), timeout=idle_timeout)
                else:
                    stdout, stderr = await proc.communicate(_input)
            except asyncio.exceptions.TimeoutError:
                proc.send_signal(SIGINT)
                raise

            # surface stderr
            if text:
                if stderr is not None:
                    stderr = smart_decode(stderr)
                if stdout is not None:
                    stdout = smart_decode(stdout)
            if proc.returncode:
                if check:
                    raise CalledProcessError(proc.returncode, command, output=stdout, stderr=stderr)
                if stderr and log_stderr:
                    command_str = " ".join(command)
                    log.warning(f"Stderr for run({command_str}):\n\t{stderr}")

            return CompletedProcess(command, proc.returncode, stdout, stderr)
        finally:
            proc_tracker.remove(proc)


async def run_live(self, *command, check=False, text=True, idle_timeout=None, **kwargs):
    """Runs a command asynchronously and iterates through its output line by line in realtime.

    This method is useful for executing a command and capturing its output on-the-fly, as it is generated.
    If an error occurs during execution, it can optionally raise an error or just log the stderr.

    Args:
        *command (str): The command to run as separate arguments.
        check (bool, optional): If set to True, raises an error if the subprocess exits with a non-zero status.
                                Defaults to False.
        text (bool, optional): If set to True, decodes the subprocess output to string. Defaults to True.
        idle_timeout (int, optional): Sets a limit on the number of seconds the process can remain idle (no lines sent to stdout) before throwing a TimeoutError
        **kwargs (dict): Additional keyword arguments for the subprocess.

    Yields:
        str or bytes: The output lines of the command, either as a decoded string (if `text=True`)
                      or as bytes (if `text=False`).

    Raises:
        CalledProcessError: If the subprocess exits with a non-zero status and `check=True`.

    Examples:
        >>> async for line in run_live(["tail", "-f", "/var/log/auth.log"]):
        ...     log.info(line)
    """
    # proc_tracker optionally keeps track of which processes are running under which modules
    # this allows for graceful SIGINTing of a module's processes in the case when it's killed
    proc_tracker = kwargs.pop("_proc_tracker", set())
    log_stderr = kwargs.pop("_log_stderr", True)
    proc, _input, command = await self._spawn_proc(*command, **kwargs)
    if proc is not None:
        proc_tracker.add(proc)
        try:
            input_task = None
            if _input is not None:
                input_task = asyncio.create_task(_write_stdin(proc, _input))

            while 1:
                try:
                    if idle_timeout is not None:
                        line = await asyncio.wait_for(proc.stdout.readline(), timeout=idle_timeout)
                    else:
                        line = await proc.stdout.readline()
                except asyncio.exceptions.TimeoutError:
                    proc.send_signal(SIGINT)
                    raise
                except ValueError as e:
                    command_str = " ".join(command)
                    log.warning(f"Error executing command {command_str}: {e}")
                    log.trace(traceback.format_exc())
                    continue
                if not line:
                    break
                if text:
                    line = smart_decode(line).rstrip("\r\n")
                else:
                    line = line.rstrip(b"\r\n")
                yield line

            if input_task is not None:
                try:
                    await input_task
                except ConnectionError:
                    log.trace(f"ConnectionError in command: {command}, kwargs={kwargs}")
                    log.trace(traceback.format_exc())
            await proc.wait()

            if proc.returncode:
                stdout, stderr = await proc.communicate()
                if text:
                    if stderr is not None:
                        stderr = smart_decode(stderr)
                    if stdout is not None:
                        stdout = smart_decode(stdout)
                if check:
                    raise CalledProcessError(proc.returncode, command, output=stdout, stderr=stderr)
                # surface stderr
                if stderr and log_stderr:
                    command_str = " ".join(command)
                    log.warning(f"Stderr for run_live({command_str}):\n\t{stderr}")
        finally:
            proc_tracker.remove(proc)


async def _spawn_proc(self, *command, **kwargs):
    """Spawns an asynchronous subprocess.

    Prepares the command and associated keyword arguments. If the `input` argument is provided,
    it checks to ensure that the `stdin` argument is not also provided. Once prepared, it creates
    and returns the subprocess. If the command executable is not found, it logs a warning and traceback.

    Args:
        *command (str): The command to run as separate arguments.
        **kwargs (dict): Additional keyword arguments for the subprocess.

    Raises:
        ValueError: If both stdin and input arguments are provided.

    Returns:
        tuple: A tuple containing the created process (or None if creation failed), the input (or None if not provided),
               and the prepared command (or None if subprocess creation failed).

    Examples:
        >>> _spawn_proc("ls", "-l", input="data")
        (<Process ...>, "data", ["ls", "-l"])
    """
    try:
        command, kwargs = self._prepare_command_kwargs(command, kwargs)
    except SubprocessError as e:
        command_str = " ".join([str(s) for s in command])
        log.warning(f"Error running command: '{command_str}': {e}")
        log.trace(traceback.format_exc())
        return None, None, None
    _input = kwargs.pop("input", None)
    if _input is not None:
        if kwargs.get("stdin") is not None:
            raise ValueError("stdin and input arguments may not both be used.")
        kwargs["stdin"] = asyncio.subprocess.PIPE

    log.hugeverbose(f"run: {' '.join(command)}")
    try:
        proc = await asyncio.create_subprocess_exec(*command, **kwargs)
        return proc, _input, command
    except FileNotFoundError as e:
        log.warning(f"{e} - missing executable?")
        log.trace(traceback.format_exc())
    return None, None, None


async def _write_proc_line(proc, chunk):
    try:
        proc.stdin.write(smart_encode(chunk) + b"\n")
        await proc.stdin.drain()
        return True
    except Exception as e:
        proc_args = [str(s) for s in getattr(proc, "args", [])]
        command = " ".join(proc_args).strip()
        if command:
            log.warning(f"Error writing line to stdin for command: {command}: {e}")
            log.trace(traceback.format_exc())
        return False


async def _write_stdin(proc, _input):
    """
    Asynchronously writes input to an active subprocess's stdin.

    This function takes an `_input` parameter, which can be of type str, bytes,
    list, tuple, or an asynchronous generator. The input is then written line by
    line to the stdin of the given `proc`.

    Args:
        proc (subprocess.Popen): An active subprocess object.
        _input (str, bytes, list, tuple, async generator): The data to write to stdin.
    """
    if _input is not None:
        if isinstance(_input, (str, bytes)):
            _input = [_input]
        if isinstance(_input, (list, tuple)):
            for chunk in _input:
                write_result = await _write_proc_line(proc, chunk)
                if not write_result:
                    break
        else:
            async for chunk in _input:
                write_result = await _write_proc_line(proc, chunk)
                if not write_result:
                    break
        proc.stdin.close()


def _prepare_command_kwargs(self, command, kwargs):
    """
    Prepare arguments for passing into `asyncio.create_subprocess_exec()`.

    This method modifies the `kwargs` dictionary in place to prepare it for
    use in the `asyncio.create_subprocess_exec()` method. It sets the default
    values for keys like 'limit', 'stdout', and 'stderr' if they are not
    already present. It also handles the case when 'sudo' needs to be run.

    Args:
        command (list): The command to be run in the subprocess.
        kwargs (dict): The keyword arguments to be passed to `asyncio.create_subprocess_exec()`.

    Returns:
        tuple: A tuple containing the modified `command` and `kwargs`.

    Examples:
        >>> _prepare_command_kwargs(['ls', '-l'], {})
        (['ls', '-l'], {'limit': 104857600, 'stdout': -1, 'stderr': -1})

        >>> _prepare_command_kwargs(['ls', '-l'], {'sudo': True})
        (['sudo', '-E', '-A', 'LD_LIBRARY_PATH=...', 'PATH=...', 'ls', '-l'], {'limit': 104857600, 'stdout': -1, 'stderr': -1, 'env': environ(...)})
    """
    # limit = 100MB (this is needed for cases like httpx that are sending large JSON blobs over stdout)
    if "limit" not in kwargs:
        kwargs["limit"] = 1024 * 1024 * 100
    if "stdout" not in kwargs:
        kwargs["stdout"] = asyncio.subprocess.PIPE
    if "stderr" not in kwargs:
        kwargs["stderr"] = asyncio.subprocess.PIPE
    sudo = kwargs.pop("sudo", False)

    if len(command) == 1 and isinstance(command[0], (list, tuple)):
        command = command[0]

    command = [str(s) for s in command]

    if not command:
        raise SubprocessError("Must specify a command")

    # use full path of binary, if not already specified
    binary = command[0]
    if "/" not in binary:
        binary_full_path = which(binary)
        if binary_full_path is None:
            raise SubprocessError(f'Command "{binary}" was not found')
        command[0] = binary_full_path

    env = kwargs.get("env", os.environ)
    if sudo and os.geteuid() != 0:
        self.depsinstaller.ensure_root()
        env["SUDO_ASKPASS"] = str((self.tools_dir / self.depsinstaller.askpass_filename).resolve())
        env["BBOT_SUDO_PASS"] = self.depsinstaller.encrypted_sudo_pw
        kwargs["env"] = env

        PATH = os.environ.get("PATH", "")
        LD_LIBRARY_PATH = os.environ.get("LD_LIBRARY_PATH", "")
        command = ["sudo", "-E", "-A", f"LD_LIBRARY_PATH={LD_LIBRARY_PATH}", f"PATH={PATH}"] + command
    return command, kwargs
